
/*--------------------------------------------------------*/

/**
 * @file
 * @author Ravi Reddy Manumachu <ravi.manumachu@ucd.ie>
 * @version 1.0
 */

/*--------------------------------------------------------*/

#ifndef _HCL_DEBUG_HPP_
#define _HCL_DEBUG_HPP_

/*--------------------------------------------------------*/

#include <string>

/*--------------------------------------------------------*/

namespace hcl {

/*-----------------------------------------------------------*/

/**
 * An class for debug information.
 */
class Debug {
public:

    /**
     * The size of the debug buffer.
     */
    static unsigned int bufferLength;

    /**
     * The verbosity level.
     */
    static unsigned int sVerbosity;

    /**
     * The HCL STREAM Library string.
     */
    static std::string sHclLibraryString;

    /**
     * Set the verbosity level.
     *
     * @param verbosity The verbosity level to set.
     */
    static void setVerbosity(
       const unsigned int verbosity
    );

    /**
     * Get the current verbosity level.
     *
     * @return The current verbosity level.
     */
    static unsigned int 
    getVerbosity();

    /**
     * Print warning messages.
     *
     * @param format The printf-style format string.
     */
    static void Warning(const char* format, ...);

    /**
     * Print error messages.
     *
     * @param format The printf-style format string.
     */
    static void Error(const char* format, ...);

    /**
     * Print info messages.
     *
     * @param format The printf-style format string.
     */
    static void Info(const char* format, ...);
};

/*--------------------------------------------------------*/

}

/*--------------------------------------------------------*/

#endif /* _HCL_DEBUG_HPP_ */

/*--------------------------------------------------------*/

