
/*-------------------------------------------------------------------------*/

/*
 * @file
 * @author Ravi Reddy Manumachu <ravi.manumachu@ucd.ie>
 * @version 1.0
 */

/*-----------------------------------------------------------*/

#include <cstdlib>
#include <cstdio>
#include <cstdarg>
#include <string>
#include <iostream>
#include <debug.hpp>

/*-----------------------------------------------------------*/

namespace hcl {

/*-----------------------------------------------------------*/

unsigned int Debug::bufferLength = 2048;

unsigned int Debug::sVerbosity = 0;

std::string Debug::sHclLibraryString = "HCL GPU GEMM OOC===> ";

/*-----------------------------------------------------------*/

void Debug::setVerbosity(
   const unsigned int verbosity
)
{
   sVerbosity = verbosity;
}

/*-----------------------------------------------------------*/

unsigned int Debug::getVerbosity()
{
   return sVerbosity;
}

/*-----------------------------------------------------------*/

void Debug::Warning(const char* format, ...)
{
    if (getVerbosity() == 0)
    {
       return;
    }
    
    char buffer[bufferLength];
    va_list argptr;
    va_start(argptr, format);
    vsprintf(buffer, format, argptr);
    va_end(argptr);  

    std::cout << Debug::sHclLibraryString;
    std::cout << " (WARNING) ";
    std::cout << buffer;
}

/*-----------------------------------------------------------*/

void Debug::Error(const char* format, ...)
{
    char buffer[bufferLength];
    va_list argptr;
    va_start(argptr, format);
    vsprintf(buffer, format, argptr);
    va_end(argptr);  

    std::cout << Debug::sHclLibraryString;
    std::cerr << " (ERROR) ";
    std::cerr << buffer;
    exit(EXIT_FAILURE);
}

/*-----------------------------------------------------------*/

void Debug::Info(const char* format, ...)
{
    if (getVerbosity() == 0)
    {
       return;
    }
    
    char buffer[bufferLength];
    va_list argptr;
    va_start(argptr, format);
    vsprintf(buffer, format, argptr);
    va_end(argptr);  

    std::cout << Debug::sHclLibraryString;
    std::cout << " (INFO) ";
    std::cout << buffer;
}

/*-----------------------------------------------------------*/

}

/*-----------------------------------------------------------*/

