
/*--------------------------------------------------------*/

/*
 * @file
 * @author zhongziming@gmail.com
 * @version 1.0
 */

/*--------------------------------------------------------*/

#include <cstdlib>
#include <iostream>
#include "gemminternal.hpp"
#include "helpers.hpp"

/*--------------------------------------------------------*/

namespace hcl {

/*--------------------------------------------------------*/

template <typename T>
int
create(
    const int M,
    const int N,
    const int K,
    const unsigned int start,
    const unsigned int end,
    const unsigned int deviceId,
    const unsigned int numDevices,
    const bool overlap,
    const unsigned int algorithm,
    gemmHandle<T>** oHandle
)
{
    gemmHandle<T>* handle = new gemmHandle<T>();

    handle->dA = NULL;
    handle->dB = NULL;
    handle->dC = NULL;
    handle->pC = NULL;
    handle->workspace_WA = NULL;
    handle->workspace_WB = NULL;
    handle->dma_WB = NULL;

    handle->idx_outer = start;
    handle->num_iteration = end;
    handle->device_id = deviceId;
    handle->num_devices = numDevices;
    handle->is_overlap = overlap;
    handle->algorithm = algorithm;

    if (handle->num_devices == 1)
    {
       if (handle->idx_outer == 0)
       {
          size_t coreDeviceMemSize = coreDeviceMemSize = OOC_GAMMA * getDeviceMemory(handle->device_id) / sizeof(T);

          handle->core_device_mem_size = coreDeviceMemSize;

          if ((M * N + M * K + K * N) > coreDeviceMemSize)
          {
             handle->kernel_type = EXCEED;
          }
          else
          {
             handle->kernel_type = FIT;
          }

          handle->num_device_use = 1;
          handle->handles = (cublasHandle_t*)malloc(sizeof(cublasHandle_t));
          CUDA_SAFE_CALL(cudaSetDevice(handle->device_id));
          CUBLAS_SAFE_CALL(cublasCreate((cublasHandle_t*)&handle->handles[0]));
       }
    }
    else
    {
       int nd = handle->num_devices;
       if (handle->idx_outer == 0)
       {
          size_t coreDeviceMemSize = coreDeviceMemSize = OOC_GAMMA * getDeviceMemory(handle->device_id) / sizeof(T);

          handle->core_device_mem_size = coreDeviceMemSize;

          if ((M * N + M * K * nd + K * N) > coreDeviceMemSize)
          {
             handle->kernel_type = EXCEED;
          }
          else
          {
             handle->kernel_type = FIT;
          }

          handle->num_device_use = nd;
          handle->handles = (cublasHandle_t*)malloc(sizeof(cublasHandle_t) * handle->num_device_use);
          
          int device, device_id = handle->device_id;
          for (device = 0; device < handle->num_device_use; device++)
          {
              CUDA_SAFE_CALL(cudaSetDevice(device_id++));
              CUBLAS_SAFE_CALL(cublasCreate((cublasHandle_t*)&handle->handles[device]));
          }
       }
    }

    *oHandle = handle;

    return 0;
}

/*--------------------------------------------------------*/

template <typename T>
int
destroy(
    gemmHandle<T>* handle
)
{
    if (handle->num_devices == 1)
    {
       CUBLAS_SAFE_CALL(cublasDestroy(handle->handles[0]));
    }
    else
    {
       int device, device_id = handle->device_id;
       for (device = 0; device < handle->num_device_use; device++)
       {
           CUBLAS_SAFE_CALL(cublasDestroy(handle->handles[device]));
       }
    }

    free(handle->handles);

    delete handle;

    return 0;
}

/*--------------------------------------------------------*/

template
struct gemmHandle<double>;

template
int
create<double>(
    const int M,
    const int N,
    const int K,
    const unsigned int start,
    const unsigned int end,
    const unsigned int deviceId,
    const unsigned int numDevices,
    const bool overlap,
    const unsigned int algorithm,
    gemmHandle<double>** oHandle
);

template
int
destroy<double>(
    gemmHandle<double>* handle
);

/*--------------------------------------------------------*/

template
struct gemmHandle<float>;

template
int
create<float>(
    const int M,
    const int N,
    const int K,
    const unsigned int start,
    const unsigned int end,
    const unsigned int deviceId,
    const unsigned int numDevices,
    const bool overlap,
    const unsigned int algorithm,
    gemmHandle<float>** oHandle
);

template
int
destroy<float>(
    gemmHandle<float>* handle
);

/*--------------------------------------------------------*/

template
struct gemmHandle<cuComplex>;

template
int
create<cuComplex>(
    const int M,
    const int N,
    const int K,
    const unsigned int start,
    const unsigned int end,
    const unsigned int deviceId,
    const unsigned int numDevices,
    const bool overlap,
    const unsigned int algorithm,
    gemmHandle<cuComplex>** oHandle
);

template
int
destroy<cuComplex>(
    gemmHandle<cuComplex>* handle
);

/*--------------------------------------------------------*/

template
struct gemmHandle<cuDoubleComplex>;

template
int
create<cuDoubleComplex>(
    const int M,
    const int N,
    const int K,
    const unsigned int start,
    const unsigned int end,
    const unsigned int deviceId,
    const unsigned int numDevices,
    const bool overlap,
    const unsigned int algorithm,
    gemmHandle<cuDoubleComplex>** oHandle
);

template
int
destroy<cuDoubleComplex>(
    gemmHandle<cuDoubleComplex>* handle
);

/*--------------------------------------------------------*/

template
struct gemmHandle<__half>;

template
int
create<__half>(
    const int M,
    const int N,
    const int K,
    const unsigned int start,
    const unsigned int end,
    const unsigned int deviceId,
    const unsigned int numDevices,
    const bool overlap,
    const unsigned int algorithm,
    gemmHandle<__half>** oHandle
);

template
int
destroy<__half>(
    gemmHandle<__half>* handle
);

/*--------------------------------------------------------*/

template <typename T>
cublasStatus_t cublasGemm(cublasHandle_t handle,
                          cublasOperation_t transa,
                          cublasOperation_t transb,
                          int m, int n, int k,
                          const T *alpha,
                          const T *A, int lda,
                          const T *B, int ldb,
                          const T *beta,
                          T *C, int ldc)
{
    std::cerr << "Unimplemented." << std::endl;
    return CUBLAS_STATUS_SUCCESS;
}

/*--------------------------------------------------------*/

template <>
cublasStatus_t cublasGemm<double>(
                          cublasHandle_t handle,
                          cublasOperation_t transa,
                          cublasOperation_t transb,
                          int m, int n, int k,
                          const double *alpha,
                          const double *A, int lda,
                          const double *B, int ldb,
                          const double *beta,
                          double *C, int ldc)
{
    return cublasDgemm(
               handle, transa, transb,
               m, n, k,
               alpha,
               A, lda,
               B, ldb,
               beta,
               C, ldc);
}

/*--------------------------------------------------------*/

template <>
cublasStatus_t cublasGemm<float>(
                          cublasHandle_t handle,
                          cublasOperation_t transa,
                          cublasOperation_t transb,
                          int m, int n, int k,
                          const float *alpha,
                          const float *A, int lda,
                          const float *B, int ldb,
                          const float *beta,
                          float *C, int ldc)
{
    return cublasSgemm(
               handle, transa, transb,
               m, n, k,
               alpha,
               A, lda,
               B, ldb,
               beta,
               C, ldc);
}

/*--------------------------------------------------------*/

template <>
cublasStatus_t cublasGemm<cuComplex>(
                          cublasHandle_t handle,
                          cublasOperation_t transa,
                          cublasOperation_t transb,
                          int m, int n, int k,
                          const cuComplex *alpha,
                          const cuComplex *A, int lda,
                          const cuComplex *B, int ldb,
                          const cuComplex *beta,
                          cuComplex *C, int ldc)
{
    return cublasCgemm(
               handle, transa, transb,
               m, n, k,
               alpha,
               A, lda,
               B, ldb,
               beta,
               C, ldc);
}

/*--------------------------------------------------------*/

template <>
cublasStatus_t cublasGemm<__half>(
                          cublasHandle_t handle,
                          cublasOperation_t transa,
                          cublasOperation_t transb,
                          int m, int n, int k,
                          const __half *alpha,
                          const __half *A, int lda,
                          const __half *B, int ldb,
                          const __half *beta,
                          __half *C, int ldc)
{
    return cublasHgemm(
               handle, transa, transb,
               m, n, k,
               alpha,
               A, lda,
               B, ldb,
               beta,
               C, ldc);
}

/*--------------------------------------------------------*/

template <>
cublasStatus_t cublasGemm<cuDoubleComplex>(
                          cublasHandle_t handle,
                          cublasOperation_t transa,
                          cublasOperation_t transb,
                          int m, int n, int k,
                          const cuDoubleComplex *alpha,
                          const cuDoubleComplex *A, int lda,
                          const cuDoubleComplex *B, int ldb,
                          const cuDoubleComplex *beta,
                          cuDoubleComplex *C, int ldc)
{
    return cublasZgemm(
               handle, transa, transb,
               m, n, k,
               alpha,
               A, lda,
               B, ldb,
               beta,
               C, ldc);
}

/*--------------------------------------------------------*/

}

/*--------------------------------------------------------*/
